@@ -8,9 +8,10 @@ use Scalar::Util qw( blessed );
 use File::Basename;
 use Carp qw(croak carp);
 use WWW::Mechanize::Link;
+use IO::Socket::INET;
 
 use vars qw($VERSION %link_spec);
-$VERSION= '0.06';
+$VERSION= '0.07';
 
 =head1 NAME
 
@@ -49,7 +50,7 @@ Specify the port where PhantomJS should listen
 
 Specify the log level of PhantomJS
 
-  log => 'OFF'
+  log => 'OFF'   # Also INFO, WARN, DEBUG
 
 =item B<launch_exe>
 
@@ -64,13 +65,24 @@ to launch. The default is the file distributed with this module.
 
 =item B<launch_arg>
 
-Specify additional parameters to the PhantomJS executable.
+Specify additional parameters to the PhantomJS executable.  (phantomjs -h)
 
   launch_arg => [ "--some-new-parameter=foo" ],
+  "--webdriver=$port",
+  '--webdriver-logfile=/tmp/webdriver',
+  '--webdriver-loglevel=DEBUG',
+  '--debug=true',
+
+  note: these set config.xxx values in ghostrdriver/config.js
 
 =item B<driver>
 
-A premade L<Selenium::Driver::Remote> object.
+A premade L<Selenium::Remote::Driver> object.
+
+=item B<report_js_errors>
+
+If set to 1, after each request tests for Javascript errors and warns. Useful
+for testing with C<use warnings qw(fatal)>.
 
 =back
 
@@ -79,7 +91,20 @@ A premade L<Selenium::Driver::Remote> object.
 sub new {
     my ($class, %options) = @_;
 
-    $options{ port } ||= 8910;
+    unless ( defined $options{ port } ) {
+        my $port = 8910;
+        while (1) {
+            $port++, next unless IO::Socket::INET->new(
+	        Listen    => 5,
+	        Proto     => 'tcp',
+	        Reuse     => 1,
+	        LocalPort => $port
+	    );
+	    last;
+        }
+    	$options{ port } = $port;
+    }
+
     $options{ "log" } ||= 'OFF';
 
     if (! exists $options{ autodie }) { $options{ autodie } = 1 };
@@ -94,26 +119,52 @@ sub new {
     $ghostdir_default= File::Spec->catfile( $ghostdir_default, 'ghostdriver', 'main.js' );
     $options{ launch_ghostdir } ||= $ghostdir_default;
     $options{ launch_arg } ||= [];
-    push @{ $options{ launch_arg }}, "--PhantomJS=$options{ port }";
+
+    # config.js defaults config.port to 8910
+    # this is the proper way to overwrite it (not sure wtf the PhantomJS parameter does above)
+    if ($options{port}) {  push @{ $options{ launch_arg }}, "--port=$options{ port }";  }  # PhantomJS version 1.9.7
+
     push @{ $options{ launch_arg }}, "--logLevel=\U$options{ log }";
     my $cmd= "| $options{ launch_exe } $options{ launch_ghostdir } @{ $options{ launch_arg } }";
-    #warn $cmd;
-    $options{ pid } ||= open my $fh, $cmd
-        or die "Couldn't launch [$cmd]: $! / $?";
-    sleep 2; # Just to give PhantomJS time to start up
-    $options{ fh } = $fh;
+    unless ($options{pid}) {
+    	$options{ kill_pid } = 1;
+    	$options{ pid } = open my $fh, $cmd
+    	    or die "Couldn't launch [$cmd]: $! / $?";
+    	$options{ fh } = $fh;
+
+        # Just to give PhantomJS time to start up, make sure it accepts connections
+        my $wait = time + ($options{ wait } || 20);
+        while ( time < $wait ) {
+            my $t = time;
+            my $socket = IO::Socket::INET->new(
+                PeerHost => '127.0.0.1',
+                PeerPort => $options{ port },
+                Proto    => 'tcp',
+            );
+            last if $socket;
+            sleep 1 if time - $t < 1;
+        }
+    }
 
     # Connect to it
-    $options{ driver } ||= Selenium::Remote::Driver->new(
-        'port' => $options{ port },
-        auto_close => 0,
-     );
-     # We patched Ghostdriver to return data, but we need to educate
-     # Selenium::Driver::Remote about it:
-     $options{ driver }->commands->get_cmds->{get}->{no_content_success}= 0;
+    eval {
+    	$options{ driver } ||= Selenium::Remote::Driver->new(
+    	    'port' => $options{ port },
+    	    auto_close => 0,
+    	);
+        # (Monkey)patch Selenium::Remote::Driver
+        $options{ driver }->commands->get_cmds->{get}->{no_content_success}= 0;
+    };
+
+    # if PhantomJS started, but so slow or unresponsive that SRD cannot connect to it,
+    # kill it manually to avoid waiting for it indefinitely
+    if ( $@ ) {
+    	kill 9, $options{ pid } if $options{ kill_pid };
+        die $@;
+    }
 
      my $self= bless \%options => $class;
-     
+
      $self->eval_in_phantomjs(<<'JS');
          var page= this;
          page.errors= [];
@@ -122,7 +173,7 @@ sub new {
              page.errors.push({ "message": msg, "trace": trace });
          };
 JS
-     
+
      $self
 };
 
@@ -163,7 +214,7 @@ sub ghostdriver_version {
     my $selenium= $mech->driver
 
 Access the L<Selenium::Driver::Remote> instance connecting to PhantomJS.
-    
+
 =cut
 
 sub driver {
@@ -253,6 +304,8 @@ sub eval_in_page {
     local @Selenium::Remote::Driver::CARP_NOT
         = (@Selenium::Remote::Driver::CARP_NOT, (ref $self)); # we trust this
     my $eval_in_sandbox = $self->driver->execute_script("return $str", @args);
+    $self->post_process;
+    return $eval_in_sandbox;
 };
 *eval = \&eval_in_page;
 
@@ -289,7 +342,7 @@ sub agent {
     my($self, $ua) = @_;
     # page.settings.userAgent = 'Mozilla/5.0 (Windows NT 5.1; rv:8.0) Gecko/20100101 Firefox/7.0';
     $self->eval_in_phantomjs(<<'JS', $ua);
-       this.settings.userAgent= arguments[0] 
+       this.settings.userAgent= arguments[0]
 JS
 }
 
@@ -374,6 +427,22 @@ C<< no_cache >> - if true, bypass the browser cache
 sub update_response {
     my( $self, $phantom_res ) = @_;
 
+    # just 1 means success
+    $phantom_res = {
+        status     => 200,
+        statusText => 'OK',
+        headers    => [{
+            name  => 'x-www-mechanize-phantomjs-fake-success',
+            value => 1,
+        }],
+    } if ref($phantom_res) eq '' and $phantom_res eq '1';
+
+    # Now add a status code of 4xx if we don't have one.
+    if( ! $phantom_res->{status}) {
+        $phantom_res->{status}= 400;
+        $phantom_res->{statusText}= "Unknown error (added by " . __PACKAGE__ . ")";
+    };
+
     my @headers= map {;@{$_}{qw(name value)}} @{ $phantom_res->{headers} };
     my $res= HTTP::Response->new( $phantom_res->{status}, $phantom_res->{statusText}, \@headers );
 
@@ -381,13 +450,15 @@ sub update_response {
 
     delete $self->{ current_form };
 
-    $self->{response} = $res
+    $self->{response} = $res;
+    return $res
 };
 
 sub get {
     my ($self, $url, %options ) = @_;
     # We need to stringify $url so it can pass through JSON
     my $phantom_res= $self->driver->get( "$url" );
+    $self->post_process;
     $self->update_response( $phantom_res );
 };
 
@@ -420,12 +491,15 @@ sub get_local {
                  Cwd::getcwd(),
              );
     $fn =~ s!\\!/!g; # fakey "make file:// URL"
-    my $url= "file:/$fn";
-
+    my $url;
+    if( $^O =~ /mswin/i ) {
+        $url= "file:///$fn";
+    } else {
+        $url= "file://$fn";
+    };
     my $res= $self->get($url, %options);
-
     # PhantomJS is not helpful with its error messages for local URLs
-    if( 0+$res->headers->header_field_names) {
+    if( 0+$res->headers->header_field_names and ([$res->headers->header_field_names]->[0] ne 'x-www-mechanize-phantomjs-fake-success' or $self->uri ne 'about:blank')) {
         # We need to fake the content headers from <meta> tags too...
         # Maybe this even needs to go into ->get()
         $res->code( 200 );
@@ -546,7 +620,7 @@ sub add_header {
     my ($self, @headers) = @_;
     use Data::Dumper;
     #warn Dumper $headers;
-    
+
     while( my ($k,$v) = splice @headers, 0, 2 ) {
         $self->eval_in_phantomjs(<<'JS', , $k, $v);
             var h= this.customHeaders;
@@ -559,7 +633,7 @@ JS
 =head2 C<< $mech->delete_header( $name , $name2... ) >>
 
     $mech->delete_header( 'User-Agent' );
-    
+
 Removes HTTP headers from the agent's list of special headers. Note
 that PhantomJS may still send a header with its default value.
 
@@ -567,7 +641,7 @@ that PhantomJS may still send a header with its default value.
 
 sub delete_header {
     my ($self, @headers) = @_;
-    
+
     $self->eval_in_phantomjs(<<'JS', @headers);
         var headers= this.customHeaders;
         for( var i = 0; i < arguments.length; i++ ) {
@@ -773,7 +847,7 @@ HTML, $mech will die.
 
 sub text {
     my $self = shift;
-    
+
     # Waugh - this is highly inefficient but conveniently short to write
     # Maybe this should skip SCRIPT nodes...
     join '', map { $_->get_text() } $self->xpath('//*/text()');
@@ -1213,7 +1287,7 @@ sub find_all_links {
 =head2 C<< $mech->find_all_links_dom %options >>
 
   print $_->{innerHTML} . "\n"
-      for $mech->find_link_dom( text_regex => qr/google/i );
+      for $mech->find_all_links_dom( text_regex => qr/google/i );
 
 Finds all matching linky DOM nodes in the document.
 The options are documented in C<< ->find_link_dom >>.
@@ -1289,7 +1363,7 @@ sub activate_container {
         warn sprintf "Switching during path to %s %s", $el->get_tag_name, $el->get_attribute('src');
         $driver->switch_to_frame( $el );
     };
-    
+
     if( ! $just_parent ) {
         warn sprintf "Activating container %s too", $doc->{id};
         # Now, unless it's the root frame, activate the container. The root frame
@@ -1453,10 +1527,10 @@ sub xpath {
         warn sprintf "Document %s", $options{ document }->{id};
     };
     #my $original_frame= $self->current_frame;
-    
+
     DOCUMENTS: {
         my $doc= $options{ document } || $self->document;
-        
+
         # This stores the path to this document
         $doc->{__path}||= [];
 
@@ -1472,7 +1546,7 @@ sub xpath {
 
             my $q = join "|", @$query;
             #warn $q;
-            
+
             my @found;
             # Now find the elements
             if ($options{ node }) {
@@ -1498,15 +1572,15 @@ sub xpath {
                 #$self->driver->switch_to_frame();
                 #warn $doc->get_text;
             };
-            
+
             # Remember the path to each found element
             for( @found ) {
                 # We reuse the reference here instead of copying the list. So don't modify the list.
                 $_->{__path}= $doc->{__path};
             };
-            
+
             push @res, @found;
-            
+
             # A small optimization to return if we already have enough elements
             # We can't do this on $return_first as there might be more elements
             #if( @res and $options{ return_first } and grep { $_->{resultSize} } @res ) {
@@ -1799,6 +1873,7 @@ and on calls to C<< ->submit() >> and C<< ->submit_with_fields >>.
 sub current_form {
     my( $self, %options )= @_;
     # Find the first <FORM> element from the currently active element
+    $self->form_number(1) unless $self->{current_form};
     $self->{current_form};
 }
 
@@ -2072,19 +2147,15 @@ sub get_set_value {
             #    $fields[0]->__event($ev);
             #};
 
-            if ('select' eq $tag) {
-                $self->select($fields[0], $value);
-            } else {
-                my $get= $self->PhantomJS_elementToJS();
-                my $val= quotemeta($value);
-                my $js= <<JS;
-                    var g=$get;
-                    var el=g("$fields[0]->{id}");
-                    el.value="$val";
+            my $get= $self->PhantomJS_elementToJS();
+            my $val= quotemeta($value);
+            my $js= <<JS;
+                var g=$get;
+                var el=g("$fields[0]->{id}");
+                el.value="$val";
 JS
-                $js= quotemeta($js);
-                $self->eval("eval('$js')"); # for some reason, Selenium/Ghostdriver don't like the JS as plain JS
-            };
+            $js= quotemeta($js);
+            $self->eval("eval('$js')"); # for some reason, Selenium/Ghostdriver don't like the JS as plain JS
 
             #for my $ev (@$post) {
             #    $fields[0]->__event($ev);
@@ -2138,6 +2209,7 @@ sub submit {
     } else {
         croak "I don't know which form to submit, sorry.";
     }
+    return $self->response;
 };
 
 =head2 C<< $mech->submit_form( %options ) >>
@@ -2188,7 +2260,7 @@ will be ignored.
 
 sub submit_form {
     my ($self,%options) = @_;
-    
+
     my $form = delete $options{ form };
     my $fields;
     if (! $form) {
@@ -2204,13 +2276,22 @@ sub submit_form {
             $form = $self->current_form;
         };
     };
-    
+
     if (! $form) {
         $self->signal_condition("No form found to submit.");
         return
     };
     $self->do_set_fields( form => $form, fields => $fields );
-    $self->submit($form);
+
+    my $response;
+    if ( $options{button} ) {
+        $response = $self->click( $options{button}, $options{x} || 0, $options{y} || 0 );
+    }
+    else {
+        $response = $self->submit();
+    }
+    return $response;
+
 }
 
 =head2 C<< $mech->set_fields( $name => $value, ... ) >>
@@ -2241,14 +2322,14 @@ sub do_set_fields {
     my ($self, %options) = @_;
     my $form = delete $options{ form };
     my $fields = delete $options{ fields };
-    
+
     while (my($n,$v) = each %$fields) {
         if (ref $v) {
             ($v,my $num) = @$v;
             warn "Index larger than 1 not supported, ignoring"
                 unless $num == 1;
         };
-        
+
         $self->get_set_value( node => $form, name => $n, value => $v, %options );
     }
 };
@@ -2270,12 +2351,12 @@ sub expand_frames {
     $spec ||= $self->{frames};
     my @spec = ref $spec ? @$spec : $spec;
     $document ||= $self->document;
-    
+
     if (! ref $spec and $spec !~ /\D/ and $spec == 1) {
         # All frames
         @spec = qw( frame iframe );
     };
-    
+
     # Optimize the default case of only names in @spec
     my @res;
     if (! grep {ref} @spec) {
@@ -2285,7 +2366,7 @@ sub expand_frames {
                         frames => 0, # otherwise we'll recurse :)
                     );
     } else {
-        @res = 
+        @res =
             map { #warn "Expanding $_";
                     ref $_
                   ? $_
@@ -2293,7 +2374,7 @@ sub expand_frames {
                   : $self->expand_frames( $_, $document );
             } @spec;
     }
-    
+
     @res
 };
 
@@ -2302,7 +2383,7 @@ sub expand_frames {
 
     my $last_frame= $mech->current_frame;
     # Switch frame somewhere else
-    
+
     # Switch back
     $mech->activate_container( $last_frame );
 
@@ -2323,7 +2404,7 @@ sub current_frame {
     my @res;
     my $current= $self->make_WebElement( $self->eval('window'));
     warn sprintf "Current_frame: bottom: %s", $current->{id};
-    
+
     # Now climb up until the root window
     my $f= $current;
     my @chain;
@@ -2331,7 +2412,7 @@ sub current_frame {
     while( $f= $self->driver->execute_script('return arguments[0].frameElement', $f )) {
         $f= $self->make_WebElement( $f );
         unshift @res, $f;
-        warn sprintf "One more level up, now in %s", 
+        warn sprintf "One more level up, now in %s",
             $f->{id};
         warn $self->driver->execute_script('return arguments[0].title', $res[0]);
         unshift @chain,
@@ -2343,8 +2424,8 @@ sub current_frame {
     warn "Chain complete";
     warn $_
         for @chain;
-    
-    # Now fake the element into 
+
+    # Now fake the element into
     my $el= $self->make_WebElement( $current );
     for( @res ) {
         warn sprintf "Path has (web element) id %s", $_->{id};
@@ -2377,68 +2458,49 @@ sub make_WebElement {
     my $res= Selenium::Remote::WebElement->new( $e->{WINDOW} || $e->{ELEMENT}, $self->driver );
     croak "No id in " . Dumper $res
         unless $res->{id};
-    
+
     $res
 }
 
 =head1 CONTENT RENDERING METHODS
 
-=head2 C<< $mech->content_as_png( [$tab, \%coordinates, \%target_size ] ) >>
+=head2 C<< $mech->content_as_png( [\%coordinates ] ) >>
 
     my $png_data = $mech->content_as_png();
 
     # Create scaled-down 480px wide preview
-    my $png_data = $mech->content_as_png(undef, undef, { width => 480 });
+    my $png_data = $mech->content_as_png(undef, { width => 480 });
 
 Returns the given tab or the current page rendered as PNG image.
 
-All parameters are optional. 
+All parameters are optional.
 
 =over 4
 
-=item *
-
-C<$tab> defaults to the current tab.
-
-=item *
+=item C< \%coordinates >
 
 If the coordinates are given, that rectangle will be cut out.
 The coordinates should be a hash with the four usual entries,
 C<left>,C<top>,C<width>,C<height>.
 
-=item *
-
-The target size of the image can also be given. It defaults to
-the size of the image. The allowed parameters in the hash are
-
-C<scalex>, C<scaley> - for specifying the scale, default is 1.0 in each direction.
-
-C<width>, C<height> - for specifying the target size
-
-If you want the resulting image to be 480 pixels wide, specify
-
-    { width => 480 }
-
-The height will then be calculated from the ratio of original width to
-original height.
-
 =back
 
 This method is specific to WWW::Mechanize::PhantomJS.
 
 Currently, the data transfer between PhantomJS and Perl
 is done Base64-encoded.
+
 =cut
 
 sub content_as_png {
-    my ($self, $tab, $rect, $target_rect) = @_;
-    #$tab ||= $self->tab;
+    my ($self, $rect) = @_;
     $rect ||= {};
-    $target_rect ||= {};
-    
-    #require MIME::Base64;
-    #my $png_base64 = $self->driver->screenshot();
-    #return MIME::Base64::decode_base64($png_base64);
+
+    if( scalar keys %$rect ) {
+
+        $self->viewport_size( $ rect );
+    };
+
     return $self->render_content( format => 'png' );
 };
 
@@ -2453,7 +2515,7 @@ Returns (or sets) the new size of the viewport (the "window").
 
 sub viewport_size {
     my( $self, $new )= @_;
-    
+
     $self->eval_in_phantomjs( <<'JS', $new );
         var viewportSize= this.clipRect;
         if( arguments[0]) {
@@ -2573,7 +2635,7 @@ sub render_content {
     my $outname= $options{ filename };
     my $format= $options{ format };
     my $wantresult;
-    
+
     my @delete;
     if( ! $outname) {
         require File::Temp;
@@ -2584,7 +2646,7 @@ sub render_content {
     };
     require File::Spec;
     $outname= File::Spec->rel2abs($outname, '.');
-    
+
     $self->eval_in_phantomjs(<<'JS', $outname, $format);
         var outname= arguments[0];
         var format= arguments[1];
@@ -2599,7 +2661,7 @@ JS
         local $/;
         $result= <$fh>;
     };
-    
+
     for( @delete ) {
         unlink $_
             or warn "Couldn't clean up tempfile: $_': $!";
@@ -2627,7 +2689,7 @@ the C<filename> option may be faster.
 
 sub content_as_pdf {
     my ($self, %options) = @_;
-    
+
     return $self->render_content( format => 'pdf', %options );
 };
 
@@ -2682,6 +2744,26 @@ sub PhantomJS_elementToJS {
 JS
 }
 
+sub post_process
+{
+    my $self = shift;
+    if ( $self->{report_js_errors} ) {
+        if ( my @errors = $self->js_errors ) {
+            $self->report_js_errors(@errors);
+            $self->clear_js_errors;
+        }
+    }
+}
+
+sub report_js_errors
+{
+    my ( $self, @errors ) = @_;
+    @errors = map {
+    	"$_->{message} at $_->{trace}->[-1]->{file} line $_->{trace}->[-1]->{line}" .
+	( $_->{trace}->[-1]->{function} ? " in function $_->{trace}->[-1]->{function}" : '')
+    } @errors;
+    Carp::carp("javascript error: @errors") ;
+}
 
 1;
 
@@ -2820,7 +2902,7 @@ L<WWW::Mechanize::Firefox> - a similar module with a visible application
 
 =head1 REPOSITORY
 
-The public repository of this module is 
+The public repository of this module is
 L<http://github.com/Corion/www-mechanize-phantomjs>.
 
 =head1 SUPPORT
